function [GammaHatGLS,GammaVarGLS]=GLS(Y,X,Z,V,Type,ConsT,ConsN)
% Input
% y = n*T by 1 vector of dependent variable
% X = n*T by 1 vector of covariate for which test of zero constraint is
% desired
% Z = n*T by k matrix of covariates other than X (if a constant or fixed effect is estimated
% it needs to be included in Z, unless it is in X)
% V = n*T by m matrix of regressors used to compute residuals for
% innovation variance estimation. If V=[] then V=[X Z] is used
%
% all data are assumed to be stored according to the following pattern: if
% y_it is the dependent variable for unit i at time t and
% y_t=(y_1t,...,y_nt)' then y=(y_1',...,y_T')'. X,Z and V are assumed to be
% organized in the same way. See the paper by Hausman and Kuersteiner for
% more detail
%
% The model estimated is of the form Y = gamma*X+beta*Z+e
% Omega = n*T by n*T GLS weight matrix defined as E(e*e')
% The code allows for Omega=SigmaT.*.eye(N)+eye(T).*.Sigma(N) where .*. is the Kronecker
% product, SigmaT is a T by T matrix with element t,s equal to cov(e_it,e_is) 
% and SigmaN is a N by N matrix with element i,j equal to cov(e_it,e_jt)
%
% ConsT is a T by T matrix of type logical. If ConsT(t,s)==true then
% cov(e_it,e_is) is estimated without constraint over the cross-section. 
% If ConsT(t,s)==false then cov(e_it,e_is)=0 
%
% ConsN is a N by N matrix of type logical. If ConsN(i,j)==true then
% cov(e_it,e_jt) is estimated without constraint over the T dimension. 
% If ConsN(t,s)==false then cov(e_it,e_is)=0 
% 
% Type ='FD' First difference the data before estimation
%      ='Fixed' Estimate the model in levels, estimate fixed effects
%      ='Level' Estimate the model in levels without fixed effects
%      
%
% Output
% GammaHatGLS = GLS parameter estimate of coefficient on X in model
% Y = gamma*X+beta*Z+e 
% 

n=length(ConsN);
T=length(ConsT);

if isempty(V)
   V=[X,Z];
   M_V=eye(T*n)-V*inv(V'*V)*V';
else
   M_V=eye(T*n)-V*inv(V'*V)*V';
end

if strcmp(Type,'FD') %first differences
      R=[-eye(T-1) zeros(T-1,1)]+[zeros(T-1,1) eye(T-1)];
      [SigmaHatT,SigmaHatN]=Comp_SHT(M_V,Y,n,T,ConsT,ConsN);                          
      kronR=kron(R,eye(n));
      SigmaHatT=R*SigmaHatT*R';
      Z=kronR*Z;
      Z=removeLinDep(Z);
      X=kronR*X;
      Y=kronR*Y;
      T=T-1;
  elseif strcmp(Type,'Fixed') 
      MT=eye(T)-ones(T,1)*ones(1,T)/T;
      B=eye(T);
      B=B(2:T,:);
      % This is an alternative method that corrects for fixed
      % effects AFTER the matrix is estimated\
      [SigmaHatT,SigmaHatN]=Comp_SHT(M_V,Y,n,T,ConsT,ConsN);
      SigmaHatT=B*MT*SigmaHatT*MT'*B';
      kronBMT=kron(B*MT,eye(n));
      Z=kronBMT*Z;
      Z=removeLinDep(Z);
      X=kronBMT*X;
      Y=kronBMT*Y;
      T=T-1;
elseif strcmp(Type,'Level') 
    [SigmaHatT,SigmaHatN]=Comp_SHT(M_V,Y,n,T,ConsT,ConsN);
else
    error('Option Type not supported');
end

OmegaT=kron(SigmaHatT,eye(n));
OmegaN=kron(eye(T),SigmaHatN);
indT=(OmegaT~=0);
indN=(OmegaN~=0);
overlap=indT.*indN;
if n>T
    indN=logical(indN-overlap);
    OmegaNstore=OmegaN;
    OmegaN=zeros(size(OmegaN));
    OmegaN(indN)=OmegaNstore(indN); %eliminate overlapping estimates in OmegaN
else
    indT=logical(indT-overlap);
    OmegaTstore=OmegaT;
    OmegaT=zeros(size(OmegaT));
    OmegaT(indT)=OmegaTstore(indT); %eliminate overlapping estimates in OmegaT
end
    

Omega=OmegaT+OmegaN;

Omegai=inv(Omega);
OZ=Omegai*Z;
Omega_z=Omegai-OZ*inv(Z'*OZ)*OZ';
XO=X'*Omega_z;
XOX=XO*X;
GammaHatGLS=XO*Y/(XOX);
GammaVarGLS=1/(XOX);
return